'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.ocpn.variants import extensive
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.ocpn.obj import OCPetriNet, OCMarking
from typing import Optional, Dict, Any
from pm4py.objects.ocel.obj import OCEL  


class Variants(Enum):
    EXTENSIVE = extensive


DEFAULT_VARIANT = Variants.EXTENSIVE
VERSIONS = {Variants.EXTENSIVE}


def apply(ocpn: OCPetriNet, initial_marking: OCMarking, final_marking: OCMarking, parameters: Optional[Dict[Any, Any]] = None, variant=DEFAULT_VARIANT) -> OCEL:
    """
    Do the playout of an object-centric Petri net generating an OCEL 2.0.

    Parameters
    -----------
    ocpn
        Object-centric Petri net to play-out
    initial_marking
        Initial marking of the object-centric Petri net
    final_marking
        Final marking of the object-centric Petri net
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm to use:
            - Variants.EXTENSIVE: gets all the traces from the model. Can be expensive
            
    Returns
    -----------
    OCEL
        Object-centric event log generated by the playout of the object-centric Petri net
    """
    return exec_utils.get_variant(variant).apply(ocpn, initial_marking, final_marking=final_marking,
                                                 parameters=parameters)
