'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocpn.variants import to_oc_causal_net, to_alternative_format
from pm4py.util import exec_utils
from enum import Enum

class Variants(Enum):
    TO_OC_CAUSAL_NET = to_oc_causal_net
    TO_ALTERNATIVE_FORMAT =  to_alternative_format
    
def apply(ocpn, parameters=None, variant=Variants.TO_OC_CAUSAL_NET):
    """
    Method for converting an Object-centric Petri Net to Object-centric Causal Net

    Parameters
    -----------
    ocpn: OCPetriNet
        Object-centric Petri net
    parameters: dict, optional
        Parameters of the algorithm
    variant
        Chosen variant of the algorithm

    Returns
    -----------
    Conversion result
    """
    return exec_utils.get_variant(variant).apply(ocpn, parameters=parameters)