'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.oc_causal_net.variants import extensive
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.oc_causal_net.obj import OCCausalNet
from typing import Optional, Dict, Any
from pm4py.objects.ocel.obj import OCEL  


class Variants(Enum):
    EXTENSIVE = extensive


DEFAULT_VARIANT = Variants.EXTENSIVE
VERSIONS = {Variants.EXTENSIVE}


def apply(occn: OCCausalNet, objects, parameters: Optional[Dict[Any, Any]] = None, variant=DEFAULT_VARIANT) -> OCEL:
    """
    Do the playout of an object-centric causal net generating an OCEL.

    Parameters
    -----------
    occn
        Object-centric causal net to play-out
    objects
        Dictionary mapping object types to object ids. These objects will be introduced by the start activities of the occn at the beginning of every binding sequence.
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm to use:
            - Variants.EXTENSIVE: gets all the traces from the model. can be expensive
            
    Returns
    -----------
    OCEL
        Object-centric event log generated by the playout of the object-centric causal net
    """
    return exec_utils.get_variant(variant).apply(occn, objects, parameters=parameters)
